from e2eml.full_processing.postprocessing import FullPipeline
from e2eml.full_processing.cpu_processing_nlp import NlpPreprocessing
import logging


class PreprocessingBluePrint(FullPipeline, NlpPreprocessing):
    def pp_bp01_std_preprocessing(self, df=None, preprocessing_type='full'):
        """
        Our recommended blueprint for model testing.
        Runs a preprocessing blueprint only. This is useful for building custom pipelines.
        :param df: Accepts a dataframe to run ml preprocessing on it.
        :param preprocessing_type: Select the type of preprocessing pipeline. "Minimum" executes the least possible steps,
        "full" the whole standard preprocessing and "nlp" adds functionality especially for NLP tasks.
        :return: Updates class attributes.
        """
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.datetime_converter(datetime_handling='all')
        self.pos_tagging_pca(pca_pos_tags=True)
        if preprocessing_type == 'nlp':
            self.append_text_sentiment_score()
            self.pos_tagging_pca(pca_pos_tags=False)
            self.tfidf_vectorizer_to_pca(pca_pos_tags=True)
            self.tfidf_naive_bayes_proba(analyzer="char_wb", ngram_range=(1, 2))
            self.tfidf_naive_bayes_proba(analyzer="word", ngram_range=(1, 1))
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.cardinality_remover(threshold=100)
        self.onehot_pca()
        self.category_encoding(algorithm='target')
        self.delete_high_null_cols(threshold=0.5)
        self.fill_nulls(how='static')
        self.data_binning(nb_bins=10)
        #self.skewness_removal()
        self.outlier_care(method='isolation', how='append')
        self.remove_collinearity(threshold=0.8)
        try:
            self.clustering_as_a_feature(algorithm='dbscan', eps=0.3, n_jobs=-1, min_samples=10)
        except ValueError:
            print("Clustering as a feature skipped due to ValueError.")
        for nb_cluster in range(2, 10):
            try:
                self.clustering_as_a_feature(algorithm='kmeans', nb_clusters=nb_cluster)
            except ValueError:
                print("Clustering as a feature skipped due to ValueError.")
        if self.low_memory_mode:
            self.reduce_memory_footprint()
        self.automated_feature_selection()
        self.sort_columns_alphabetically()

    def pp_bp02_std_preprocessing(self, df=None, preprocessing_type='full'):
        """
        This preprocessing blueprint contains alternative decision compare to pp_bp01.
        Runs a preprocessing blueprint only. This is useful for building custom pipelines.
        :param df: Accepts a dataframe to run ml preprocessing on it.
        :param preprocessing_type: Select the type of preprocessing pipeline. "Minimum" executes the least possible steps,
        "full" the whole standard preprocessing and "nlp" adds functionality especially for NLP tasks.
        :return: Updates class attributes.
        """
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.datetime_converter(datetime_handling='all')
        self.pos_tagging_pca(pca_pos_tags=True)
        if preprocessing_type == 'nlp':
            self.pos_tagging_pca(pca_pos_tags=False)
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.cardinality_remover(threshold=100)
        self.onehot_pca()
        self.category_encoding(algorithm='GLMM')
        self.delete_high_null_cols(threshold=0.4)
        self.fill_nulls(how='iterative_imputation', fill_with=0)
        self.data_binning(nb_bins=20)
        #self.skewness_removal()
        self.outlier_care(method='isolation', how='append')
        self.remove_collinearity(threshold=0.85)
        try:
            self.clustering_as_a_feature(algorithm='dbscan', eps=0.3, n_jobs=-1, min_samples=10)
        except ValueError:
            print("Clustering as a feature skipped due to ValueError.")
        for nb_cluster in range(2, 10):
            try:
                self.clustering_as_a_feature(algorithm='GLMM', nb_clusters=nb_cluster)
            except ValueError:
                print("Clustering as a feature skipped due to ValueError.")
        if self.low_memory_mode:
            self.reduce_memory_footprint()
        self.automated_feature_selection()
        self.sort_columns_alphabetically()

    def pp_bp03_std_preprocessing(self, df=None, preprocessing_type='full'):
        """
        This blueprint adds skewness removal by log transformation, data scaling and SMOTE.
        Runs a preprocessing blueprint only. This is useful for building custom pipelines.
        :param df: Accepts a dataframe to run ml preprocessing on it.
        :param preprocessing_type: Select the type of preprocessing pipeline. "Minimum" executes the least possible steps,
        "full" the whole standard preprocessing and "nlp" adds functionality especially for NLP tasks.
        :return: Updates class attributes.
        """
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.datetime_converter(datetime_handling='all')
        self.pos_tagging_pca(pca_pos_tags=True)
        if preprocessing_type == 'nlp':
            self.pos_tagging_pca(pca_pos_tags=False)
            #self.tfidf_vectorizer_to_pca(pca_pos_tags=True)
            self.tfidf_naive_bayes_proba()
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.cardinality_remover(threshold=100)
        self.onehot_pca()
        self.category_encoding(algorithm='GLMM')
        self.delete_high_null_cols(threshold=0.4)
        self.fill_nulls(how='static', fill_with=-99)
        self.data_binning(nb_bins=5)
        self.skewness_removal()
        self.outlier_care(method='isolation', how='append')
        self.remove_collinearity(threshold=0.80)
        try:
            self.clustering_as_a_feature(algorithm='dbscan', eps=0.3, n_jobs=-1, min_samples=10)
        except ValueError:
            print("Clustering as a feature skipped due to ValueError.")
        for nb_cluster in range(2, 10):
            try:
                self.clustering_as_a_feature(algorithm='GLMM', nb_clusters=nb_cluster)
            except ValueError:
                print("Clustering as a feature skipped due to ValueError.")
        if self.low_memory_mode:
            self.reduce_memory_footprint()
        self.automated_feature_selection()
        self.sort_columns_alphabetically()
        self.data_scaling()
        if self.class_problem == 'binary' or self.class_problem == 'multiclass':
            self.smote_data()
        else:
            pass

    def pp_bp10_nlp_preprocessing(self, df):
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.sort_columns_alphabetically()
        self.check_max_sentence_length()
        self.import_transformer_model_tokenizer(transformer_chosen=self.transformer_chosen)

    def pp_bp11_nlp_preprocessing(self, df):
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.regex_clean_text_data()
        self.sort_columns_alphabetically()
        self.check_max_sentence_length()
        self.import_transformer_model_tokenizer(transformer_chosen=self.transformer_chosen)

    def pp_bp12_nlp_preprocessing(self, df):
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.oversample_train_data()
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.sort_columns_alphabetically()
        self.check_max_sentence_length()
        self.import_transformer_model_tokenizer(transformer_chosen=self.transformer_chosen)

    def pp_bp13_nlp_preprocessing(self, df):
        logging.info('Start blueprint.')
        self.runtime_warnings(warn_about="future_architecture_change")
        try:
            if df.empty:
                self.prediction_mode = False
            else:
                self.dataframe = df
                self.prediction_mode = True
        except AttributeError:
            self.prediction_mode = False
        self.train_test_split(how=self.train_split_type)
        self.replace_synonyms_to_df_copy(words_to_replace=3, mode='auto')
        #self.oversample_train_data()
        self.rare_feature_processor(threshold=0.005, mask_as='miscellaneous', rarity_cols=self.rarity_cols)
        self.sort_columns_alphabetically()
        self.check_max_sentence_length()
        self.import_transformer_model_tokenizer(transformer_chosen=self.transformer_chosen)



from e2eml.full_processing import postprocessing
from statsmodels.tsa.stattools import adfuller
from statsmodels.tsa.stattools import kpss
import matplotlib.pyplot as plt
import pandas as pd

plt.style.use('fivethirtyeight')


class TimeSeriesPreprocessing(postprocessing.FullPipeline):
    def time_series_unpacking(self):
        """
        Takes the datasource dictionary within the class and unpacks it, depending on the provided source format.
        :return: Returns unpacked data objects.
        """
        if self.source_format == 'numpy array':
            Y_train, Y_test = self.np_array_unpack_test_train_dict()
            return Y_train, Y_test
        elif self.source_format == 'Pandas dataframe':
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            return X_train, X_test, Y_train, Y_test

    def time_series_wrap_test_train_to_dict(self, X_train=None, X_test=None, Y_train=None, Y_test=None):
        """
        Takes either four Pandas dataframes and series or two numpy arrays and packs them into the class
        dictionary.
        :param X_train: Pandas dataframe (optional)
        :param X_test: Pandas dataframe (optional)
        :param Y_train: Pandas series or numpy array
        :param Y_test: Pandas series or numpy array
        :return: Updates class and return dictionary.
        """
        if self.source_format == 'numpy array':
            return self.np_array_wrap_test_train_to_dict(Y_train, Y_test)
        elif self.source_format == 'Pandas dataframe':
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def augmented_dickey_fuller_test(self, window):
        """
        The Dickey Fuller test is one of the most popular statistical tests. It can be used to determine the
        presence of unit root in the series, and hence help us understand if the series is stationary or not.
        The null and alternate hypothesis of this test are:
          - Null Hypothesis: The series has a unit root (value of a =1)
          - Alternate Hypothesis: The series has no unit root.
        If we fail to reject the null hypothesis, we can say that the series is non-stationary.
        This means that the series can be linear or difference stationary.
        :return:
        """
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        # Determing rolling statistics
        rolmean = Y_train.rolling(window=window).mean()
        rolstd = Y_train.rolling(window=window).std()
        # Plot rolling statistics:
        plt.plot(Y_train, color='blue', label='Original')
        plt.plot(rolmean, color='red', label='Rolling Mean')
        plt.plot(rolstd, color='black', label='Rolling Std')
        plt.legend(loc='best')
        plt.title('Rolling Mean & Standard Deviation')
        plt.show()
        # Perform Dickey-Fuller test:
        print('Results of Dickey-Fuller Test:')
        dftest = adfuller(Y_train, autolag='AIC')
        dfoutput = pd.Series(dftest[0:4],
                             index=['Test Statistic', 'p-value', '#Lags Used', 'Number of Observations Used'])
        for key, value in dftest[4].items():
            dfoutput['Critical Value (%s)' % key] = value
        print(dfoutput)

    def kwiatkowski_phillips_schmidt_shin_test(self):
        """
        KPSS is another test for checking the stationarity of a time series (slightly less popular than the Dickey
        Fuller test). The null and alternate hypothesis for the KPSS test are opposite that of the ADF test,
        which often creates confusion.
        The authors of the KPSS test have defined the null hypothesis as the process is trend stationary,
        to an alternate hypothesis of a unit root series.
        - Null Hypothesis: The process is trend stationary.
        - Alternate Hypothesis: The series has a unit root (series is not stationary).
        Test for stationarity: If the test statistic is greater than the critical value,
        we reject the null hypothesis (series is not stationary). If the test statistic is less than the critical value,
         if fail to reject the null hypothesis (series is stationary).
        So in summary, the ADF test has an alternate hypothesis of linear or difference stationary,
        while the KPSS test identifies trend-stationarity in a series.
        :return:
        """
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        print('Results of KPSS Test:')
        kpsstest = kpss(Y_train, regression='c')
        kpss_output = pd.Series(kpsstest[0:3], index=['Test Statistic', 'p-value', 'Lags Used'])
        for key, value in kpsstest[3].items():
            kpss_output['Critical Value (%s)' % key] = value
        print(kpss_output)

    def stationarity_explainer(self):
        explanation = """
        What are the different types of stationarity?
        - Strict Stationary: A strict stationary series satisfies the mathematical definition of a stationary process. 
          For a strict stationary series, the mean, variance and covariance are not the function of time. 
          The aim is to convert a non-stationary series into a strict stationary series for making predictions.
        - Trend Stationary: A series that has no unit root but exhibits a trend is referred to as a trend stationary 
          series. Once the trend is removed, the resulting series will be strict stationary. The KPSS test classifies a 
          series as stationary on the absence of unit root. This means that the series can be strict stationary or 
          trend stationary.
        - Difference Stationary: A time series that can be made strict stationary by differencing falls under difference 
          stationary. ADF test is also known as a difference stationarity test.

        It’s always better to apply both the tests, so that we are sure that the series is truly stationary. 
        Let us look at the possible outcomes of applying these stationary tests.

        - Case 1: Both tests conclude that the series is not stationary -> series is not stationary
        - Case 2: Both tests conclude that the series is stationary -> series is stationary
        - Case 3: KPSS = stationary and ADF = not stationary -> trend stationary, remove the trend to make series 
          strict stationary
        - Case 4: KPSS = not stationary and ADF = stationary -> difference stationary, use differencing to make series 
          stationary
        """
        print(explanation)



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["babel>=2.9.0",
                "boostaroota>=1.3",
                "category_encoders==2.2.2",
                "dill>=0.3.3",
                "imblearn>=0.0",
                "lightgbm==3.1.0",
                "matplotlib>=3.3.4",
                "ngboost>=0.3.1",
                "nltk>=3.2.4",
                "numpy>=1.19.4",
                "optuna>=2.5.0",
                "pandas==1.1.5",
                "pip>=21.0.1",
                "plotly>=4.14.3",
                "psutil==5.8.0",
                "seaborn>=0.11.1",
                "scikit-learn==0.22.2",
                "scipy==1.6.3",
                "setuptools >= 51.1.0",
                "shap>=0.39.0",
                "spacy>=2.3.0",
                "textblob>=0.15.3",
                "torch >= 1.7.0",
                "transformers>=4.0.0",
                "vowpalwabbit>=8.11.0",
                "wheel==0.36.2",
                "xgboost>=1.3.3",
                ]
extras_require = {
    'rapids': ['cupy', 'cython'],
    'jupyter': ['ipython>=7.10.0', 'jupyter_core>=4.7.0'],
    'full': ['cupy', 'cython', 'ipython>=7.10.0', 'jupyter_core>=4.7.0']
}

setup(
    name="e2eml",
    version="1.6.5",
    author="Thomas Meißner",
    author_email="meissnercorporation@gmx.de",
    description="An end to end solution for automl.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/ThomasMeissnerDS/e2e_ml",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)